//---------------------------------------------------------------------------
#include "Sound.h"
#include "Resource.h"

//---------------------------------------------------------------------------

Sound::Sound() {
	this->initialised = false;
}

//#ifdef V_USE_SDL

#ifdef _WIN32
#include <SDL.h>
#include <SDL_mixer.h>
#elif __linux
#include <SDL/SDL.h>
#include <SDL/SDL_mixer.h>
#endif

SDL_Sound::SDL_Sound() : Sound() {
	for(int i=0;i<max_sounds_c;i++) {
		chunks[i] = NULL;
		channels[i] = -1;
	}

	if( SDL_InitSubSystem(SDL_INIT_AUDIO) != 0 ) {
		LOG("failed to init SDL audio subsystem");
		return;
	}
	if( Mix_OpenAudio(44100, MIX_DEFAULT_FORMAT, 2, 1024) == -1 ) {
		LOG("Mix_OpenAudio failed: %s\n", Mix_GetError());
		SDL_QuitSubSystem(SDL_INIT_AUDIO);
		return;
	}
#ifdef _WIN32
    // not yet available on Linux!
	Mix_Init(0);
#endif
	this->initialised = true;
}

SDL_Sound::~SDL_Sound() {
	for(int i=0;i<max_sounds_c;i++) {
		if( chunks[i] != NULL )
			Mix_FreeChunk(chunks[i]);
	}
	if( this->initialised ) {
		Mix_CloseAudio();
#ifdef _WIN32
        // not yet available on Linux!
		Mix_Quit();
#endif
		SDL_QuitSubSystem(SDL_INIT_AUDIO);
	}
}

bool SDL_Sound::LoadSound(int s,const char *file) {
	ASSERT( s >= 0 && s < max_sounds_c );
	if( !initialised )
		return false;
	chunks[s] = Mix_LoadWAV(file);
	if( chunks[s] == NULL )
		return false;
	return true;
}

bool SDL_Sound::PlaySound(int s,bool loop,int delay) {
	ASSERT( s >= 0 && s < max_sounds_c );
	if( !initialised )
		return false;
	if( chunks[s] == NULL )
		return false;
	if( channels[s] != -1 ) {
		if( Mix_Paused(channels[s]) ) {
			// sound was paused, so let's just resume
			Mix_Resume(channels[s]);
			return true;
		}
		// otherwise, let's stop the currently playing sound
		StopSound(s, true);
	}
	channels[s] = Mix_PlayChannel(-1, chunks[s], loop ? -1 : 0);
	if( channels[s] == -1 ) {
		LOG("Failed to play sound %d: %s\n", s, Mix_GetError());
		return false;
	}
	return true;
}

void SDL_Sound::FreeSound(int s) {
	ASSERT( s >= 0 && s < max_sounds_c );
	if( !initialised )
		return;
	if( chunks[s] == NULL )
		return;

	if( channels[s] != -1 )
		StopSound(s, true);
	Mix_FreeChunk(chunks[s]);
}

bool SDL_Sound::IsSoundPlaying(int s) {
	ASSERT( s >= 0 && s < max_sounds_c );
	if( !initialised )
		return false;
	if( chunks[s] == NULL )
		return false;
	if( channels[s] == -1 )
		return false;
	if( Mix_Paused(channels[s]) != 0 )
		return false; // paused, so count as not playing
	return true;
}

void SDL_Sound::StopSound(int s,bool bResetPosition) {
	ASSERT( s >= 0 && s < max_sounds_c );
	if( !initialised )
		return;
	if( chunks[s] == NULL )
		return;
	if( channels[s] == -1 )
		return;
	if( bResetPosition ) {
		Mix_HaltChannel(channels[s]);
		channels[s] = -1;
	}
	else {
		Mix_Pause(channels[s]);
		// n.b., don't reset channels[s], so we can still unpause it
	}
}

void SDL_Sound::playMedia(const char *filename) {
	// TODO
}

bool SDL_Sound::isMediaPlaying() {
	// TODO
	return false;
}

//#endif
