#pragma once

/** Represents a Civilization, together with AI routines. Also stores Race, which defines particular properties
*   of a Civilization that do not change during the game.
*/

#include "../Vision/VisionIface.h"

#include "buildable.h" // for Age enum

#include <vector>
using std::vector;
#include <string>
using std::string;

class MainGamestate;
class Civilization;
class City;
class Technology;
class Unit;

class Race {
	string name;
	string name_adj;
	unsigned char rgb[3];
	VI_Texture *texture;
	vector<string> city_names;
	const Technology *starting_technology;
public:
	Race(const char *name, const char *name_adj, unsigned char rgb[3], const char *image_filename);
	virtual ~Race();

	const char *getName() const {
		return this->name.c_str();
	}
	const char *getNameAdjective() const {
		return this->name_adj.c_str();
	}
	const unsigned char *getColor() const {
		return this->rgb;
	}
	VI_Texture *getTexture() const {
		// n.b., return non-const needed due to draw functions being non-const
		return this->texture;
	}
	void setStartingTechnology(const Technology *starting_technology) {
		this->starting_technology = starting_technology;
	}
	const Technology *getStartingTechnology() const {
		return this->starting_technology;
	}
	bool equals(const char *name) const {
		bool equal = this->name == name;
		return equal;
	}
	void addCityName(const char *city_name) {
		this->city_names.push_back(city_name);
	}
	string makeCityName(int *looped, int *index) const;
	void advanceCityName(int *looped, int *index) const;
};

class Relationship {
	MainGamestate *mainGamestate;
	Civilization *civ1, *civ2;
public:
	// n.b., enum Status fails to compile on Linux, when including X11 headers!
	enum RelationshipStatus {
		STATUS_PEACE = 0,
		STATUS_WAR = 1
	};
protected:
	bool made_contact;
	RelationshipStatus status;
	int year_declared_war;

	Relationship(MainGamestate *mainGamestate);
public:
	Relationship(MainGamestate *mainGamestate, Civilization *civ1, Civilization *civ2);

	static Relationship *load(MainGamestate *mainGamestate, FILE *file);
	void save(FILE *file) const;

	bool equals(const Civilization *civ1, const Civilization *civ2) const {
		if( ( this->civ1 == civ1 && this->civ2 == civ2 ) ||
			( this->civ1 == civ2 && this->civ2 == civ1 ) ) {
				return true;
		}
		return false;
	}
	bool madeContact() const {
		return made_contact;
	}
	void makeContact();
	RelationshipStatus getStatus() const {
		return status;
	}
	void setStatus(RelationshipStatus status);
	int getYearDeclaredWar() const;
};

class Civilization {
public:
	// n.b., the values of each enum must not be changed, in order to provide backwards compatibility with save game files
	enum Bonus {
		BONUS_TECH = 0,
		BONUS_MILITARY = 1,
		BONUS_GROWTH = 2,
		BONUS_PRODUCTION = 3,
		BONUS_WORKERS = 4,
		N_BONUSES = 5
	};
private:
	MainGamestate *mainGamestate;
	const Race *race; // race name saved
	vector<City *> cities; // saved
	vector<const Technology *> technologies; // saved
	vector<Unit *> units; // saved
	//bool is_dead; // TODO: save

	int progress_technology; // saved
	const Technology *technology; // saved

	int power; // saved

	int city_name_looped; // saved
	int city_name_index; // saved

	int map_width, map_height; // not saved
    bool *map_visible; // saved

	bool done_ai;

	int bonus_turns[N_BONUSES]; // saved

	void checkDead();
	void setMapVisible(int x,int y,bool visible);

	Civilization(MainGamestate *mainGamestate);
public:
	Civilization(MainGamestate *mainGamestate, const Race *race);
	~Civilization();

	static Civilization *load(MainGamestate *mainGamestate, FILE *file);
	void save(FILE *file) const;

	bool isDead() const;
	const char *getName() const {
		return this->race->getName();
	}
	const char *getNameAdjective() const {
		return this->race->getNameAdjective();
	}
	const unsigned char *getColor() const {
		return this->race->getColor();
	}
	VI_Texture *getTexture() const {
		// n.b., return non-const needed due to draw functions being non-const
		return this->race->getTexture();
	}
	string makeCityName();
	void advanceCityName();

	void addCity(City *city);
	void removeCity(const City *city);
	int getNCities() const;
	const City *getCity(int i) const;
	City *getCity(int i);

	void addUnit(Unit *unit);
	void removeUnit(const Unit *unit, bool check_dead);
	void removeUnit(const Unit *unit) {
		removeUnit(unit, true);
	}
	int getNUnits() const;
	const Unit *getUnit(int i) const;
	Unit *getUnit(int i);
	/*const vector<Unit *> *getUnits() const {
		return &units;
	}*/

	bool hasTechnology(const Technology *technology) const;
	bool hasAllTechnology() const;
	/*const vector<const Technology *> *getTechnologies() const {
		return &technologies;
	}*/
	bool canResearch(const Technology *technology) const;
	const Technology *getTechnology() const {
		return technology;
	}
	void setTechnology(const Technology *technology) {
		this->technology = technology;
	}
	Technology::Age getAge() const;
	void removeAllTechnologies();
	void addTechnology(const Technology *tech);
	bool updateProgressTechnology(int update);
	int getProgressTechnology() const {
		return progress_technology;
	}
	int getPower() const {
		return power;
	}
	int getPowerPerTurn() const;
	void updatePower(int update);
	bool hasBonus(Bonus bonus) const;

	int getTotalScience() const;
	int getTotalProduction() const;
	int getTotalPopulation() const;
	int getMilitary(int *n_units) const;
	int getAir(int *n_units) const;
	int getNuclear() const;

	void resetAI() {
		this->done_ai = false;
	}
	void setDoneAI() {
		this->done_ai = true;
	}
	bool doneAI() const {
		return done_ai;
	}
	void update();
	/*bool isDead() const {
		return is_dead;
	}
	void setDead();*/
	bool isExplored(int x,int y) const;
	bool isExploredBase(int x,int y) const;
	bool isExploredBase(int indx) const;
	bool uncover(int x,int y,int range);
	void makeContact(int x,int y,int range);
	void revealMap();
	bool compareMaps(const Civilization *that) const;
	bool canShareMaps(Civilization *that) const;
	void shareMaps(Civilization *that);

	//int getArmyStrength() const;

	static void getDistinctTechnologies(MainGamestate *mainGamestate, vector<const Technology *> *technologies1, vector<const Technology *> *technologies2, const Civilization *civ1, const Civilization *civ2);

	// AI functions
	void chooseTechnologyAI();
};
